@@ -42,7 +42,8 @@
 #define DEFAULT_INPUT_REQUEST_TIMEOUT 120 * 1000
 #define DEFAULT_BROWSER_LAUNCH_TIMEOUT 300 * 1000
 
-#define CONFIGMAINFILE CONFIGDIR "/main.conf"
+#define MAINFILE "main.conf"
+#define CONFIGMAINFILE CONFIGDIR "/" MAINFILE
 
 static char *default_auto_connect[] = {
 	"wifi",
@@ -68,6 +69,7 @@ static struct {
 	unsigned int timeout_browserlaunch;
 	char **blacklisted_interfaces;
 	connman_bool_t allow_hostname_updates;
+	connman_bool_t single_tech;
 } connman_settings  = {
 	.bg_scan = TRUE,
 	.pref_timeservers = NULL,
@@ -78,6 +80,32 @@ static struct {
 	.timeout_browserlaunch = DEFAULT_BROWSER_LAUNCH_TIMEOUT,
 	.blacklisted_interfaces = NULL,
 	.allow_hostname_updates = TRUE,
+	.single_tech = FALSE,
+};
+
+#define CONF_BG_SCAN                    "BackgroundScanning"
+#define CONF_PREF_TIMESERVERS           "FallbackTimeservers"
+#define CONF_AUTO_CONNECT               "DefaultAutoConnectTechnologies"
+#define CONF_PREFERRED_TECHS            "PreferredTechnologies"
+#define CONF_FALLBACK_NAMESERVERS       "FallbackNameservers"
+#define CONF_TIMEOUT_INPUTREQ           "InputRequestTimeout"
+#define CONF_TIMEOUT_BROWSERLAUNCH      "BrowserLaunchTimeout"
+#define CONF_BLACKLISTED_INTERFACES     "NetworkInterfaceBlacklist"
+#define CONF_ALLOW_HOSTNAME_UPDATES     "AllowHostnameUpdates"
+#define CONF_SINGLE_TECH                "SingleConnectedTechnology"
+
+static const char *supported_options[] = {
+	CONF_BG_SCAN,
+	CONF_PREF_TIMESERVERS,
+	CONF_AUTO_CONNECT,
+	CONF_PREFERRED_TECHS,
+	CONF_FALLBACK_NAMESERVERS,
+	CONF_TIMEOUT_INPUTREQ,
+	CONF_TIMEOUT_BROWSERLAUNCH,
+	CONF_BLACKLISTED_INTERFACES,
+	CONF_ALLOW_HOSTNAME_UPDATES,
+	CONF_SINGLE_TECH,
+	NULL
 };
 
 static GKeyFile *load_config(const char *file)
@@ -151,6 +179,45 @@ static char **parse_fallback_nameservers(char **nameservers, gsize len)
 	return servers;
 }
 
+static void check_config(GKeyFile *config)
+{
+	char **keys;
+	int j;
+
+	if (config == NULL)
+		return;
+
+	keys = g_key_file_get_groups(config, NULL);
+
+	for (j = 0; keys != NULL && keys[j] != NULL; j++) {
+		if (g_strcmp0(keys[j], "General") != 0)
+			connman_warn("Unknown group %s in %s",
+						keys[j], MAINFILE);
+	}
+
+	g_strfreev(keys);
+
+	keys = g_key_file_get_keys(config, "General", NULL, NULL);
+
+	for (j = 0; keys != NULL && keys[j] != NULL; j++) {
+		connman_bool_t found;
+		int i;
+
+		found = FALSE;
+		for (i = 0; supported_options[i] != NULL; i++) {
+			if (g_strcmp0(keys[j], supported_options[i]) == 0) {
+				found = TRUE;
+				break;
+			}
+		}
+		if (found == FALSE && supported_options[i] == NULL)
+			connman_warn("Unknown option %s in %s",
+						keys[j], MAINFILE);
+	}
+
+	g_strfreev(keys);
+}
+
 static void parse_config(GKeyFile *config)
 {
 	GError *error = NULL;
@@ -169,24 +236,24 @@ static void parse_config(GKeyFile *config)
 		return;
 	}
 
-	DBG("parsing main.conf");
+	DBG("parsing %s", MAINFILE);
 
 	boolean = g_key_file_get_boolean(config, "General",
-						"BackgroundScanning", &error);
+						CONF_BG_SCAN, &error);
 	if (error == NULL)
 		connman_settings.bg_scan = boolean;
 
 	g_clear_error(&error);
 
 	timeservers = g_key_file_get_string_list(config, "General",
-						"FallbackTimeservers", NULL, &error);
+					CONF_PREF_TIMESERVERS, NULL, &error);
 	if (error == NULL)
 		connman_settings.pref_timeservers = timeservers;
 
 	g_clear_error(&error);
 
 	str_list = g_key_file_get_string_list(config, "General",
-			"DefaultAutoConnectTechnologies", &len, &error);
+			CONF_AUTO_CONNECT, &len, &error);
 
 	if (error == NULL)
 		connman_settings.auto_connect =
@@ -200,7 +267,7 @@ static void parse_config(GKeyFile *config)
 	g_clear_error(&error);
 
 	str_list = g_key_file_get_string_list(config, "General",
-			"PreferredTechnologies", &len, &error);
+			CONF_PREFERRED_TECHS, &len, &error);
 
 	if (error == NULL)
 		connman_settings.preferred_techs =
@@ -211,7 +278,7 @@ static void parse_config(GKeyFile *config)
 	g_clear_error(&error);
 
 	str_list = g_key_file_get_string_list(config, "General",
-			"FallbackNameservers", &len, &error);
+			CONF_FALLBACK_NAMESERVERS, &len, &error);
 
 	if (error == NULL)
 		connman_settings.fallback_nameservers =
@@ -222,21 +289,21 @@ static void parse_config(GKeyFile *config)
 	g_clear_error(&error);
 
 	timeout = g_key_file_get_integer(config, "General",
-			"InputRequestTimeout", &error);
+			CONF_TIMEOUT_INPUTREQ, &error);
 	if (error == NULL && timeout >= 0)
 		connman_settings.timeout_inputreq = timeout * 1000;
 
 	g_clear_error(&error);
 
 	timeout = g_key_file_get_integer(config, "General",
-			"BrowserLaunchTimeout", &error);
+			CONF_TIMEOUT_BROWSERLAUNCH, &error);
 	if (error == NULL && timeout >= 0)
 		connman_settings.timeout_browserlaunch = timeout * 1000;
 
 	g_clear_error(&error);
 
 	interfaces = g_key_file_get_string_list(config, "General",
-			"NetworkInterfaceBlacklist", &len, &error);
+			CONF_BLACKLISTED_INTERFACES, &len, &error);
 
 	if (error == NULL)
 		connman_settings.blacklisted_interfaces = interfaces;
@@ -247,12 +314,19 @@ static void parse_config(GKeyFile *config)
 	g_clear_error(&error);
 
 	boolean = g_key_file_get_boolean(config, "General",
-					"AllowHostnameUpdates",
+					CONF_ALLOW_HOSTNAME_UPDATES,
 					&error);
 	if (error == NULL)
 		connman_settings.allow_hostname_updates = boolean;
 
 	g_clear_error(&error);
+
+	boolean = g_key_file_get_boolean(config, "General",
+			CONF_SINGLE_TECH, &error);
+	if (error == NULL)
+		connman_settings.single_tech = boolean;
+
+	g_clear_error(&error);
 }
 
 static int config_init(const char *file)
@@ -260,6 +334,7 @@ static int config_init(const char *file)
 	GKeyFile *config;
 
 	config = load_config(file);
+	check_config(config);
 	parse_config(config);
 	if (config != NULL)
 		g_key_file_free(config);
@@ -414,24 +489,27 @@ const char *connman_option_get_string(const char *key)
 
 connman_bool_t connman_setting_get_bool(const char *key)
 {
-	if (g_str_equal(key, "BackgroundScanning") == TRUE)
+	if (g_str_equal(key, CONF_BG_SCAN) == TRUE)
 		return connman_settings.bg_scan;
 
-	if (g_str_equal(key, "AllowHostnameUpdates") == TRUE)
+	if (g_str_equal(key, CONF_ALLOW_HOSTNAME_UPDATES) == TRUE)
 		return connman_settings.allow_hostname_updates;
 
+	if (g_str_equal(key, CONF_SINGLE_TECH) == TRUE)
+		return connman_settings.single_tech;
+
 	return FALSE;
 }
 
 char **connman_setting_get_string_list(const char *key)
 {
-	if (g_str_equal(key, "FallbackTimeservers") == TRUE)
+	if (g_str_equal(key, CONF_PREF_TIMESERVERS) == TRUE)
 		return connman_settings.pref_timeservers;
 
-	if (g_str_equal(key, "FallbackNameservers") == TRUE)
+	if (g_str_equal(key, CONF_FALLBACK_NAMESERVERS) == TRUE)
 		return connman_settings.fallback_nameservers;
 
-	if (g_str_equal(key, "NetworkInterfaceBlacklist") == TRUE)
+	if (g_str_equal(key, CONF_BLACKLISTED_INTERFACES) == TRUE)
 		return connman_settings.blacklisted_interfaces;
 
 	return NULL;
@@ -439,10 +517,10 @@ char **connman_setting_get_string_list(const char *key)
 
 unsigned int *connman_setting_get_uint_list(const char *key)
 {
-	if (g_str_equal(key, "DefaultAutoConnectTechnologies") == TRUE)
+	if (g_str_equal(key, CONF_AUTO_CONNECT) == TRUE)
 		return connman_settings.auto_connect;
 
-	if (g_str_equal(key, "PreferredTechnologies") == TRUE)
+	if (g_str_equal(key, CONF_PREFERRED_TECHS) == TRUE)
 		return connman_settings.preferred_techs;
 
 	return NULL;
@@ -535,7 +613,7 @@ int main(int argc, char *argv[])
 	g_dbus_set_disconnect_function(conn, disconnect_callback, NULL, NULL);
 
 	__connman_log_init(argv[0], option_debug, option_detach,
-			option_backtrace);
+			option_backtrace, "Connection Manager", VERSION);
 
 	__connman_dbus_init(conn);
 
@@ -545,14 +623,15 @@ int main(int argc, char *argv[])
 		config_init(option_config);
 
 	__connman_storage_migrate();
+	__connman_inotify_init();
 	__connman_technology_init();
 	__connman_notifier_init();
+	__connman_agent_init();
 	__connman_service_init();
 	__connman_provider_init();
 	__connman_network_init();
 	__connman_device_init(option_device, option_nodevice);
 
-	__connman_agent_init();
 	__connman_ippool_init();
 	__connman_iptables_init();
 	__connman_nat_init();
@@ -597,8 +676,8 @@ int main(int argc, char *argv[])
 	__connman_wpad_cleanup();
 	__connman_dhcpv6_cleanup();
 	__connman_dhcp_cleanup();
-	__connman_provider_cleanup();
 	__connman_plugin_cleanup();
+	__connman_provider_cleanup();
 	__connman_connection_cleanup();
 	__connman_timeserver_cleanup();
 	__connman_session_cleanup();
@@ -613,7 +692,6 @@ int main(int argc, char *argv[])
 	__connman_config_cleanup();
 	__connman_manager_cleanup();
 	__connman_counter_cleanup();
-	__connman_agent_cleanup();
 	__connman_tethering_cleanup();
 	__connman_nat_cleanup();
 	__connman_iptables_cleanup();
@@ -621,9 +699,11 @@ int main(int argc, char *argv[])
 	__connman_device_cleanup();
 	__connman_network_cleanup();
 	__connman_service_cleanup();
+	__connman_agent_cleanup();
 	__connman_ipconfig_cleanup();
 	__connman_notifier_cleanup();
 	__connman_technology_cleanup();
+	__connman_inotify_cleanup();
 
 	__connman_dbus_cleanup();
 
