@@ -89,6 +89,7 @@ struct connman_session {
 	struct session_info *info;
 	struct session_info *info_last;
 	struct connman_session_config *policy_config;
+	GSList *user_allowed_bearers;
 
 	connman_bool_t ecall;
 
@@ -160,7 +161,23 @@ static const char *type2string(enum connman_session_type type)
 	return NULL;
 }
 
-static enum connman_session_type string2type(const char *type)
+enum connman_session_roaming_policy connman_session_parse_roaming_policy(const char *policy)
+{
+	if (g_strcmp0(policy, "default") == 0)
+		return CONNMAN_SESSION_ROAMING_POLICY_DEFAULT;
+	else if (g_strcmp0(policy, "always") == 0)
+		return CONNMAN_SESSION_ROAMING_POLICY_ALWAYS;
+	else if (g_strcmp0(policy, "forbidden") == 0)
+		return CONNMAN_SESSION_ROAMING_POLICY_FORBIDDEN;
+	else if (g_strcmp0(policy, "national") == 0)
+		return CONNMAN_SESSION_ROAMING_POLICY_NATIONAL;
+	else if (g_strcmp0(policy, "international") == 0)
+		return CONNMAN_SESSION_ROAMING_POLICY_INTERNATIONAL;
+	else
+		return CONNMAN_SESSION_ROAMING_POLICY_UNKNOWN;
+}
+
+enum connman_session_type connman_session_parse_connection_type(const char *type)
 {
 	if (g_strcmp0(type, "any") == 0)
 		return CONNMAN_SESSION_TYPE_ANY;
@@ -172,25 +189,26 @@ static enum connman_session_type string2type(const char *type)
 	return CONNMAN_SESSION_TYPE_UNKNOWN;
 }
 
-static enum connman_service_type bearer2service(const char *bearer)
+static int bearer2service(const char *bearer, enum connman_service_type *type)
 {
-	if (bearer == NULL)
-		return CONNMAN_SERVICE_TYPE_UNKNOWN;
-
 	if (g_strcmp0(bearer, "ethernet") == 0)
-		return CONNMAN_SERVICE_TYPE_ETHERNET;
+		*type = CONNMAN_SERVICE_TYPE_ETHERNET;
 	else if (g_strcmp0(bearer, "wifi") == 0)
-		return CONNMAN_SERVICE_TYPE_WIFI;
+		*type = CONNMAN_SERVICE_TYPE_WIFI;
 	else if (g_strcmp0(bearer, "wimax") == 0)
-		return CONNMAN_SERVICE_TYPE_WIMAX;
+		*type = CONNMAN_SERVICE_TYPE_WIMAX;
 	else if (g_strcmp0(bearer, "bluetooth") == 0)
-		return CONNMAN_SERVICE_TYPE_BLUETOOTH;
+		*type = CONNMAN_SERVICE_TYPE_BLUETOOTH;
 	else if (g_strcmp0(bearer, "cellular") == 0)
-		return CONNMAN_SERVICE_TYPE_CELLULAR;
+		*type = CONNMAN_SERVICE_TYPE_CELLULAR;
 	else if (g_strcmp0(bearer, "vpn") == 0)
-		return CONNMAN_SERVICE_TYPE_VPN;
+		*type = CONNMAN_SERVICE_TYPE_VPN;
+	else if (g_strcmp0(bearer, "*") == 0)
+		*type = CONNMAN_SERVICE_TYPE_UNKNOWN;
 	else
-		return CONNMAN_SERVICE_TYPE_UNKNOWN;
+		return -EINVAL;
+
+	return 0;
 }
 
 static char *service2bearer(enum connman_service_type type)
@@ -218,50 +236,104 @@ static char *service2bearer(enum connman_service_type type)
 	return "";
 }
 
-static int assign_policy_plugin(struct connman_session *session)
-{
-	GSList *list;
-	struct connman_session_policy *policy;
-
-	if (session->policy != NULL)
-		return -EALREADY;
-
-	for (list = policy_list; list != NULL; list = list->next) {
-		policy = list->data;
-
-		session->policy = policy;
-		break;
-	}
-
-	return 0;
-}
-
-static int create_policy_config(struct connman_session *session)
-{
-	struct connman_session_config *config;
-
-	if (session->policy == NULL)
-		config = connman_session_create_default_config();
-	else
-		config = (*session->policy->create)(session);
-
-	if (config == NULL)
-		return -ENOMEM;
-
-	session->policy_config = config;
-
-	return 0;
-}
-
 static void destroy_policy_config(struct connman_session *session)
 {
 	if (session->policy == NULL) {
-		connman_session_free_bearers(
-			session->policy_config->allowed_bearers);
 		g_free(session->policy_config);
-	} else {
-		(*session->policy->destroy)(session);
+		return;
 	}
+
+	(*session->policy->destroy)(session);
+}
+
+static void free_session(struct connman_session *session)
+{
+	if (session == NULL)
+		return;
+
+	if (session->notify_watch > 0)
+		g_dbus_remove_watch(connection, session->notify_watch);
+
+	destroy_policy_config(session);
+	g_slist_free(session->info->config.allowed_bearers);
+	g_free(session->owner);
+	g_free(session->session_path);
+	g_free(session->notify_path);
+	g_free(session->info);
+	g_free(session->info_last);
+
+	g_free(session);
+}
+
+static void cleanup_session(gpointer user_data)
+{
+	struct connman_session *session = user_data;
+	struct session_info *info = session->info;
+
+	DBG("remove %s", session->session_path);
+
+	g_slist_free(session->user_allowed_bearers);
+	g_hash_table_destroy(session->service_hash);
+	g_sequence_free(session->service_list);
+
+	if (info->entry != NULL &&
+			info->entry->reason == CONNMAN_SESSION_REASON_CONNECT) {
+		__connman_service_disconnect(info->entry->service);
+	}
+
+	free_session(session);
+}
+
+static int assign_policy_plugin(struct connman_session *session)
+{
+	if (session->policy != NULL)
+		return -EALREADY;
+
+	if (policy_list == NULL)
+		return 0;
+
+	session->policy = policy_list->data;
+
+	return 0;
+}
+
+struct user_config {
+	DBusMessage *pending;
+
+	enum connman_session_type type;
+	GSList *allowed_bearers;
+};
+
+static void cleanup_user_config(struct user_config *user_config)
+{
+	if (user_config == NULL)
+		return;
+
+	if (user_config->pending != NULL)
+		dbus_message_unref(user_config->pending);
+
+	g_slist_free(user_config->allowed_bearers);
+	g_free(user_config);
+}
+
+static int create_policy_config(struct connman_session *session,
+				connman_session_config_cb callback,
+				struct user_config *user_config)
+{
+	struct connman_session_config *config;
+
+	if (session->policy == NULL) {
+		config = connman_session_create_default_config();
+		if (config == NULL) {
+			free_session(session);
+			cleanup_user_config(user_config);
+			return -ENOMEM;
+		}
+
+		return callback(session, config, user_config, 0);
+	}
+
+	return (*session->policy->create)(session, callback, user_config);
 }
 
 static void probe_policy(struct connman_session_policy *policy)
@@ -319,6 +391,9 @@ int connman_session_policy_register(struct connman_session_policy *policy)
 {
 	DBG("name %s", policy->name);
 
+	if (policy->create == NULL || policy->destroy == NULL)
+		return -EINVAL;
+
 	policy_list = g_slist_insert_sorted(policy_list, policy,
 						compare_priority);
 
@@ -336,6 +411,22 @@ void connman_session_policy_unregister(struct connman_session_policy *policy)
 	remove_policy(policy);
 }
 
+int connman_session_set_default_config(struct connman_session_config *config)
+{
+	config->priority = FALSE;
+	config->roaming_policy = CONNMAN_SESSION_ROAMING_POLICY_DEFAULT;
+	config->type = CONNMAN_SESSION_TYPE_ANY;
+	config->ecall = FALSE;
+
+	g_slist_free(config->allowed_bearers);
+	config->allowed_bearers = g_slist_prepend(NULL,
+				GINT_TO_POINTER(CONNMAN_SERVICE_TYPE_UNKNOWN));
+	if (config->allowed_bearers == NULL)
+		return -ENOMEM;
+
+	return 0;
+}
+
 struct connman_session_config *connman_session_create_default_config(void)
 {
 	struct connman_session_config *config;
@@ -344,12 +435,7 @@ struct connman_session_config *connman_session_create_default_config(void)
 	if (config == NULL)
 		return NULL;
 
-	config->priority = FALSE;
-	config->roaming_policy = CONNMAN_SESSION_ROAMING_POLICY_DEFAULT;
-	config->type = CONNMAN_SESSION_TYPE_ANY;
-	config->ecall = FALSE;
-	config->allowed_bearers = connman_session_allowed_bearers_any();
-	if (config->allowed_bearers == NULL) {
+	if (connman_session_set_default_config(config) < 0) {
 		g_free(config);
 		return NULL;
 	}
@@ -373,89 +459,76 @@ static enum connman_session_type apply_policy_on_type(
 	return CONNMAN_SESSION_TYPE_INTERNET;
 }
 
-static void cleanup_bearer(gpointer data)
+int connman_session_parse_bearers(const char *token, GSList **list)
 {
-	struct connman_session_bearer *bearer = data;
+	enum connman_service_type bearer;
+	int err;
 
-	g_free(bearer->name);
-	g_free(bearer);
+	if (g_strcmp0(token, "") == 0)
+		return 0;
+
+	err = bearer2service(token, &bearer);
+	if (err < 0)
+		return err;
+
+	*list = g_slist_append(*list, GINT_TO_POINTER(bearer));
+
+	return 0;
 }
 
-void connman_session_free_bearers(GSList *bearers)
+static int parse_bearers(DBusMessageIter *iter, GSList **list)
 {
-	g_slist_free_full(bearers, cleanup_bearer);
-}
-
-static int session_parse_allowed_bearers(DBusMessageIter *iter, GSList **list)
-{
-	struct connman_session_bearer *bearer;
 	DBusMessageIter array;
+	int type, err;
 
 	dbus_message_iter_recurse(iter, &array);
 
 	*list = NULL;
 
-	while (dbus_message_iter_get_arg_type(&array) == DBUS_TYPE_STRING) {
+	while ((type = dbus_message_iter_get_arg_type(&array)) !=
+			DBUS_TYPE_INVALID) {
 		char *bearer_name = NULL;
 
+		if (type != DBUS_TYPE_STRING) {
+			g_slist_free(*list);
+			*list = NULL;
+			return -EINVAL;
+		}
+
 		dbus_message_iter_get_basic(&array, &bearer_name);
 
-		bearer = g_try_new0(struct connman_session_bearer, 1);
-		if (bearer == NULL) {
-			connman_session_free_bearers(*list);
+		err = connman_session_parse_bearers(bearer_name, list);
+		if (err < 0) {
+			g_slist_free(*list);
 			*list = NULL;
-			return -ENOMEM;
+			return err;
 		}
 
-		bearer->name = g_strdup(bearer_name);
-		bearer->service_type = bearer2service(bearer->name);
-
-		if (bearer->service_type == CONNMAN_SERVICE_TYPE_UNKNOWN &&
-				g_strcmp0(bearer->name, "*") == 0) {
-			bearer->match_all = TRUE;
-		} else {
-			bearer->match_all = FALSE;
-		}
-
-		*list = g_slist_append(*list, bearer);
-
 		dbus_message_iter_next(&array);
 	}
 
 	return 0;
 }
 
-static struct connman_session_bearer *clone_bearer(
-					struct connman_session_bearer *orig)
-{
-	struct connman_session_bearer *bearer;
-
-	bearer = g_try_new0(struct connman_session_bearer, 1);
-	if (bearer == NULL)
-		return NULL;
-
-	bearer->name = g_strdup(orig->name);
-	bearer->match_all = orig->match_all;
-	bearer->service_type = orig->service_type;
-
-	return bearer;
-}
-
 static int filter_bearer(GSList *policy_bearers,
-				struct connman_session_bearer *bearer,
+				enum connman_service_type bearer,
 				GSList **list)
 {
-	struct connman_session_bearer *policy, *tmp;
+	enum connman_service_type policy;
 	GSList *it;
 
 	if (policy_bearers == NULL)
 		goto clone;
 
 	for (it = policy_bearers; it != NULL; it = it->next) {
-		policy = it->data;
+		policy = GPOINTER_TO_INT(it->data);
 
-		if (policy->match_all == FALSE &&
-				bearer->service_type != policy->service_type)
+		if (bearer == CONNMAN_SERVICE_TYPE_UNKNOWN) {
+			bearer = policy;
+			goto clone;
+		}
+
+		if (policy != CONNMAN_SERVICE_TYPE_UNKNOWN && policy != bearer)
 			continue;
 
 		goto clone;
@@ -466,40 +539,22 @@ static int filter_bearer(GSList *policy_bearers,
 	return 0;
 
 clone:
-	tmp = clone_bearer(bearer);
-	if (tmp == NULL) {
-		connman_session_free_bearers(*list);
-		*list = NULL;
-		return -ENOMEM;
-	}
-	*list = g_slist_append(*list, tmp);
+	*list = g_slist_append(*list, GINT_TO_POINTER(bearer));
 
 	return 0;
 }
 
-static connman_bool_t is_bearer_valid(struct connman_session_bearer *bearer)
-{
-	if (bearer->match_all == FALSE &&
-			bearer->service_type == CONNMAN_SERVICE_TYPE_UNKNOWN)
-		return FALSE;
-
-	return TRUE;
-}
-
 static int apply_policy_on_bearers(GSList *policy_bearers, GSList *bearers,
 				GSList **list)
 {
-	struct connman_session_bearer *bearer;
+	enum connman_service_type bearer;
 	GSList *it;
 	int err;
 
 	*list = NULL;
 
 	for (it = bearers; it != NULL; it = it->next) {
-		bearer = it->data;
-
-		if (is_bearer_valid(bearer) == FALSE)
-			continue;
+		bearer = GPOINTER_TO_INT(it->data);
 
 		err = filter_bearer(policy_bearers, bearer, list);
 		if (err < 0)
@@ -509,22 +564,9 @@ static int apply_policy_on_bearers(GSList *policy_bearers, GSList *bearers,
 	return 0;
 }
 
-GSList *connman_session_allowed_bearers_any(void)
+const char *connman_session_get_owner(struct connman_session *session)
 {
-	struct connman_session_bearer *bearer;
-	GSList *list = NULL;
-
-	bearer = g_try_new0(struct connman_session_bearer, 1);
-	if (bearer == NULL)
-		return NULL;
-
-	bearer->name = g_strdup("");
-	bearer->match_all = TRUE;
-	bearer->service_type = CONNMAN_SERVICE_TYPE_UNKNOWN;
-
-	list = g_slist_append(list, bearer);
-
-	return list;
+	return session->owner;
 }
 
 static void append_allowed_bearers(DBusMessageIter *iter, void *user_data)
@@ -534,10 +576,14 @@ static void append_allowed_bearers(DBusMessageIter *iter, void *user_data)
 
 	for (list = info->config.allowed_bearers;
 			list != NULL; list = list->next) {
-		struct connman_session_bearer *bearer = list->data;
+		enum connman_service_type bearer = GPOINTER_TO_INT(list->data);
+		const char *name = __connman_service_type2string(bearer);
+
+		if (name == NULL)
+			name = "*";
 
 		dbus_message_iter_append_basic(iter, DBUS_TYPE_STRING,
-						&bearer->name);
+						&name);
 	}
 }
 
@@ -761,14 +807,14 @@ static connman_bool_t service_type_match(struct connman_session *session,
 
 	for (list = info->config.allowed_bearers;
 			list != NULL; list = list->next) {
-		struct connman_session_bearer *bearer = list->data;
+		enum connman_service_type bearer = GPOINTER_TO_INT(list->data);
 		enum connman_service_type service_type;
 
-		if (bearer->match_all == TRUE)
+		if (bearer == CONNMAN_SERVICE_TYPE_UNKNOWN)
 			return TRUE;
 
 		service_type = connman_service_get_type(service);
-		if (bearer->service_type == service_type)
+		if (bearer == service_type)
 			return TRUE;
 	}
 
@@ -832,9 +878,9 @@ static gint sort_allowed_bearers(struct connman_service *service_a,
 
 	for (list = info->config.allowed_bearers;
 			list != NULL; list = list->next) {
-		struct connman_session_bearer *bearer = list->data;
+		enum connman_service_type bearer = GPOINTER_TO_INT(list->data);
 
-		if (bearer->match_all == TRUE) {
+		if (bearer == CONNMAN_SERVICE_TYPE_UNKNOWN) {
 			if (type_a != type_b) {
 				weight_a = service_type_weight(type_a);
 				weight_b = service_type_weight(type_b);
@@ -849,20 +895,14 @@ static gint sort_allowed_bearers(struct connman_service *service_a,
 			}
 		}
 
-		if (type_a == bearer->service_type &&
-				type_b == bearer->service_type) {
+		if (type_a == bearer && type_b == bearer)
 			return 0;
-		}
 
-		if (type_a == bearer->service_type &&
-				type_b != bearer->service_type) {
+		if (type_a == bearer &&	type_b != bearer)
 			return -1;
-		}
 
-		if (type_a != bearer->service_type &&
-				type_b == bearer->service_type) {
+		if (type_a != bearer &&	type_b == bearer)
 			return 1;
-		}
 	}
 
 	return 0;
@@ -878,32 +918,6 @@ static gint sort_services(gconstpointer a, gconstpointer b, gpointer user_data)
 				session);
 }
 
-static void cleanup_session(gpointer user_data)
-{
-	struct connman_session *session = user_data;
-	struct session_info *info = session->info;
-
-	DBG("remove %s", session->session_path);
-
-	g_hash_table_destroy(session->service_hash);
-	g_sequence_free(session->service_list);
-
-	if (info->entry != NULL &&
-			info->entry->reason == CONNMAN_SESSION_REASON_CONNECT) {
-		__connman_service_disconnect(info->entry->service);
-	}
-
-	destroy_policy_config(session);
-	connman_session_free_bearers(session->info->config.allowed_bearers);
-	g_free(session->owner);
-	g_free(session->session_path);
-	g_free(session->notify_path);
-	g_free(session->info);
-	g_free(session->info_last);
-
-	g_free(session);
-}
-
 static enum connman_session_state service_to_session_state(enum connman_service_state state)
 {
 	switch (state) {
@@ -1393,6 +1407,46 @@ static void session_changed(struct connman_session *session,
 	session_notify(session);
 }
 
+int connman_session_config_update(struct connman_session *session)
+{
+	struct session_info *info = session->info;
+	GSList *allowed_bearers;
+	int err;
+
+	DBG("session %p", session);
+
+	/*
+	 * We update all configuration even though only one entry
+	 * might have changed. We can still optimize this later.
+	 */
+
+	err = apply_policy_on_bearers(
+		session->policy_config->allowed_bearers,
+		session->user_allowed_bearers,
+		&allowed_bearers);
+	if (err < 0)
+		return err;
+
+	g_slist_free(info->config.allowed_bearers);
+	info->config.allowed_bearers = allowed_bearers;
+
+	info->config.type = apply_policy_on_type(
+				session->policy_config->type,
+				info->config.type);
+
+	info->config.roaming_policy = session->policy_config->roaming_policy;
+
+	info->config.ecall = session->policy_config->ecall;
+	if (info->config.ecall == TRUE)
+		ecall_session = session;
+
+	info->config.priority = session->policy_config->priority;
+
+	session_changed(session, CONNMAN_SESSION_TRIGGER_SETTING);
+
+	return 0;
+}
+
 static DBusMessage *connect_session(DBusConnection *conn,
 					DBusMessage *msg, void *user_data)
 {
@@ -1460,18 +1514,18 @@ static DBusMessage *change_session(DBusConnection *conn,
 	switch (dbus_message_iter_get_arg_type(&value)) {
 	case DBUS_TYPE_ARRAY:
 		if (g_str_equal(name, "AllowedBearers") == TRUE) {
-			err = session_parse_allowed_bearers(&value,
-							&allowed_bearers);
+			err = parse_bearers(&value, &allowed_bearers);
 			if (err < 0)
 				return __connman_error_failed(msg, err);
 
-			connman_session_free_bearers(info->config.allowed_bearers);
+			g_slist_free(info->config.allowed_bearers);
+			session->user_allowed_bearers = allowed_bearers;
+
 			err = apply_policy_on_bearers(
 					session->policy_config->allowed_bearers,
-					allowed_bearers,
+					session->user_allowed_bearers,
 					&info->config.allowed_bearers);
 
-			connman_session_free_bearers(allowed_bearers);
 			if (err < 0)
 				return __connman_error_failed(msg, err);
 		} else {
@@ -1483,7 +1537,7 @@ static DBusMessage *change_session(DBusConnection *conn,
 			dbus_message_iter_get_basic(&value, &val);
 			info->config.type = apply_policy_on_type(
 				session->policy_config->type,
-				string2type(val));
+				connman_session_parse_connection_type(val));
 		} else {
 			goto err;
 		}
@@ -1577,17 +1631,99 @@ static const GDBusMethodTable session_methods[] = {
 	{ },
 };
 
+static int session_create_cb(struct connman_session *session,
+				struct connman_session_config *config,
+				void *user_data, int err)
+{
+	DBusMessage *reply;
+	struct user_config *user_config = user_data;
+	struct session_info *info, *info_last;
+
+	DBG("session %p config %p", session, config);
+
+	if (err != 0)
+		goto out;
+
+	session->policy_config = config;
+
+	info = session->info;
+	info_last = session->info_last;
+
+	if (session->policy_config->ecall == TRUE)
+		ecall_session = session;
+
+	info->state = CONNMAN_SESSION_STATE_DISCONNECTED;
+	info->config.type = apply_policy_on_type(
+				session->policy_config->type,
+				user_config->type);
+	info->config.priority = session->policy_config->priority;
+	info->config.roaming_policy = session->policy_config->roaming_policy;
+	info->entry = NULL;
+
+	session->user_allowed_bearers = user_config->allowed_bearers;
+	user_config->allowed_bearers = NULL;
+
+	err = apply_policy_on_bearers(
+			session->policy_config->allowed_bearers,
+			session->user_allowed_bearers,
+			&info->config.allowed_bearers);
+	if (err < 0)
+		goto out;
+
+	g_hash_table_replace(session_hash, session->session_path, session);
+
+	DBG("add %s", session->session_path);
+
+	if (g_dbus_register_interface(connection, session->session_path,
+					CONNMAN_SESSION_INTERFACE,
+					session_methods, NULL,
+					NULL, session, NULL) == FALSE) {
+		connman_error("Failed to register %s", session->session_path);
+		g_hash_table_remove(session_hash, session->session_path);
+		err = -EINVAL;
+		goto out;
+	}
+
+	reply = g_dbus_create_reply(user_config->pending,
+				DBUS_TYPE_OBJECT_PATH, &session->session_path,
+				DBUS_TYPE_INVALID);
+	g_dbus_send_message(connection, reply);
+	user_config->pending = NULL;
+
+	populate_service_list(session);
+
+	info_last->state = info->state;
+	info_last->config.priority = info->config.priority;
+	info_last->config.roaming_policy = info->config.roaming_policy;
+	info_last->entry = info->entry;
+	info_last->config.allowed_bearers = info->config.allowed_bearers;
+
+	session->append_all = TRUE;
+
+	session_changed(session, CONNMAN_SESSION_TRIGGER_SETTING);
+
+out:
+	if (err < 0) {
+		reply = __connman_error_failed(user_config->pending, -err);
+		g_dbus_send_message(connection, reply);
+
+		free_session(session);
+	}
+
+	cleanup_user_config(user_config);
+
+	return err;
+}
+
 int __connman_session_create(DBusMessage *msg)
 {
 	const char *owner, *notify_path;
 	char *session_path = NULL;
 	DBusMessageIter iter, array;
 	struct connman_session *session = NULL;
-	struct session_info *info, *info_last;
-	enum connman_session_type type = CONNMAN_SESSION_TYPE_ANY;
-	GSList *allowed_bearers = NULL;
-	connman_bool_t allowed_bearers_valid = FALSE;
-	connman_bool_t type_valid = FALSE;
+	struct user_config *user_config = NULL;
+	connman_bool_t user_allowed_bearers = FALSE;
+	connman_bool_t user_connection_type = FALSE;
 	int err;
 
 	owner = dbus_message_get_sender(msg);
@@ -1603,6 +1739,14 @@ int __connman_session_create(DBusMessage *msg)
 		goto err;
 	}
 
+	user_config = g_try_new0(struct user_config, 1);
+	if (user_config == NULL) {
+		err = -ENOMEM;
+		goto err;
+	}
+
+	user_config->pending = dbus_message_ref(msg);
+
 	dbus_message_iter_init(msg, &iter);
 	dbus_message_iter_recurse(&iter, &array);
 
@@ -1619,28 +1763,52 @@ int __connman_session_create(DBusMessage *msg)
 		switch (dbus_message_iter_get_arg_type(&value)) {
 		case DBUS_TYPE_ARRAY:
 			if (g_str_equal(key, "AllowedBearers") == TRUE) {
-				err = session_parse_allowed_bearers(&value,
-							&allowed_bearers);
+				err = parse_bearers(&value,
+						&user_config->allowed_bearers);
 				if (err < 0)
 					goto err;
 
-				allowed_bearers_valid = TRUE;
+				user_allowed_bearers = TRUE;
 			} else {
-				return -EINVAL;
+				err = -EINVAL;
+				goto err;
 			}
 			break;
 		case DBUS_TYPE_STRING:
 			if (g_str_equal(key, "ConnectionType") == TRUE) {
 				dbus_message_iter_get_basic(&value, &val);
-				type = string2type(val);
-				type_valid = TRUE;
+				user_config->type =
+					connman_session_parse_connection_type(val);
+
+				user_connection_type = TRUE;
 			} else {
-				return -EINVAL;
+				err = -EINVAL;
+				goto err;
 			}
 		}
 		dbus_message_iter_next(&array);
 	}
 
+	/*
+	 * If the user hasn't provided a configuration, we set
+	 * the default configuration.
+	 *
+	 * For AllowedBearers this is '*', ...
+	 */
+	if (user_allowed_bearers == FALSE) {
+		user_config->allowed_bearers =
+			g_slist_append(NULL,
+				GINT_TO_POINTER(CONNMAN_SERVICE_TYPE_UNKNOWN));
+		if (user_config->allowed_bearers == NULL) {
+			err = -ENOMEM;
+			goto err;
+		}
+	}
+
+	/* ... and for ConnectionType it is 'any'. */
+	if (user_connection_type == FALSE)
+		user_config->type = CONNMAN_SESSION_TYPE_ANY;
+
 	dbus_message_iter_next(&iter);
 	dbus_message_iter_get_basic(&iter, &notify_path);
 
@@ -1657,6 +1825,7 @@ int __connman_session_create(DBusMessage *msg)
 
 	session = g_hash_table_lookup(session_hash, session_path);
 	if (session != NULL) {
+		g_free(session_path);
 		session = NULL;
 		err = -EEXIST;
 		goto err;
@@ -1664,10 +1833,13 @@ int __connman_session_create(DBusMessage *msg)
 
 	session = g_try_new0(struct connman_session, 1);
 	if (session == NULL) {
+		g_free(session_path);
 		err = -ENOMEM;
 		goto err;
 	}
 
+	session->session_path = session_path;
+
 	session->info = g_try_new0(struct session_info, 1);
 	if (session->info == NULL) {
 		err = -ENOMEM;
@@ -1680,104 +1852,38 @@ int __connman_session_create(DBusMessage *msg)
 		goto err;
 	}
 
-	info = session->info;
-	info_last = session->info_last;
-
 	session->owner = g_strdup(owner);
-	session->session_path = session_path;
 	session->notify_path = g_strdup(notify_path);
 	session->notify_watch =
 		g_dbus_add_disconnect_watch(connection, session->owner,
 					owner_disconnect, session, NULL);
 
 	err = assign_policy_plugin(session);
-	if (err < 0)
-		goto err;
-	err = create_policy_config(session);
 	if (err < 0)
 		goto err;
 
-	if (session->policy_config->ecall == TRUE)
-		ecall_session = session;
+	err = create_policy_config(session, session_create_cb, user_config);
+	if (err < 0 && err != -EINPROGRESS)
+		return err;
 
-	info->state = CONNMAN_SESSION_STATE_DISCONNECTED;
-	if (type_valid == FALSE)
-		type = CONNMAN_SESSION_TYPE_ANY;
-	info->config.type = apply_policy_on_type(
-				session->policy_config->type,
-				type);
-	info->config.priority = session->policy_config->priority;
-	info->config.roaming_policy = session->policy_config->roaming_policy;
-	info->entry = NULL;
-
-	if (allowed_bearers_valid == FALSE) {
-		allowed_bearers = connman_session_allowed_bearers_any();
-		if (allowed_bearers == NULL) {
-			err = -ENOMEM;
-			goto err;
-		}
-	}
-
-	err = apply_policy_on_bearers(
-			session->policy_config->allowed_bearers,
-			allowed_bearers,
-			&info->config.allowed_bearers);
-	if (err < 0)
-		goto err;
-
-	g_hash_table_replace(session_hash, session->session_path, session);
-
-	DBG("add %s", session->session_path);
-
-	if (g_dbus_register_interface(connection, session->session_path,
-					CONNMAN_SESSION_INTERFACE,
-					session_methods, NULL,
-					NULL, session, NULL) == FALSE) {
-		connman_error("Failed to register %s", session->session_path);
-		g_hash_table_remove(session_hash, session->session_path);
-		session = NULL;
-
-		err = -EINVAL;
-		goto err;
-	}
-
-	g_dbus_send_reply(connection, msg,
-				DBUS_TYPE_OBJECT_PATH, &session->session_path,
-				DBUS_TYPE_INVALID);
-
-
-	populate_service_list(session);
-
-	info_last->state = info->state;
-	info_last->config.priority = info->config.priority;
-	info_last->config.roaming_policy = info->config.roaming_policy;
-	info_last->entry = info->entry;
-	info_last->config.allowed_bearers = info->config.allowed_bearers;
-
-	session->append_all = TRUE;
-
-	session_changed(session, CONNMAN_SESSION_TRIGGER_SETTING);
-
-	return 0;
+	return -EINPROGRESS;
 
 err:
 	connman_error("Failed to create session");
 
-	if (session != NULL) {
-		if (session->info_last != NULL)
-			g_free(session->info_last);
-		if (session->info != NULL)
-			g_free(session->info);
-		g_free(session);
-	}
-
-	g_free(session_path);
-
-	connman_session_free_bearers(allowed_bearers);
+	free_session(session);
 
+	cleanup_user_config(user_config);
 	return err;
 }
 
+void connman_session_destroy(struct connman_session *session)
+{
+	DBG("session %p", session);
+
+	session_disconnect(session);
+}
+
 int __connman_session_destroy(DBusMessage *msg)
 {
 	const char *owner, *session_path;
@@ -1799,7 +1905,7 @@ int __connman_session_destroy(DBusMessage *msg)
 	if (g_strcmp0(owner, session->owner) != 0)
 		return -EACCES;
 
-	session_disconnect(session);
+	connman_session_destroy(session);
 
 	return 0;
 }
